package	jp.co.sds_corp.fixconnect.message;

import	java.io.ByteArrayOutputStream;
import	java.io.Serializable;
import	java.text.DecimalFormat;

import	jp.co.sds_corp.fixconnect.FIXException;

public class FIXMessage implements Serializable
{
	/** serialVersionUID */
    private static final long serialVersionUID = 1L;

	public static final byte SOH = 0x01;
	/** レイアウト情報 */
	Layout layout = new Layout();

	protected FIXMessage() {
	}

	public FIXMessage( String msgType ) throws FIXException {
		this();
		try {
			LayoutBuilder builder = LayoutBuilder.getInstance();
			builder.createLayout( this, msgType );
		} catch ( Exception ex ) {
			throw( new FIXException( "create layout error.", ex ));
		}
	}

	/**
	* レイアウト情報を返します.
	* @return	レイアウト情報
	*/
	public Layout getLayout() {
		return this.layout;
	}

	/**
	 * 電文データ項目を登録します.
	 * コンポーネントを含めて"COMPONENT[index].FIELD"形式でも指定可能です。
	 * @param	name	項目名
	 * @param	data	設定するデータ
	 * @throws FieldNotFoundException
	 */
	public void setField( String name, String data )
		throws FieldNotFoundException
	{
		Field field = getLayout().findField( name );
		field.setValue( data );
	}

	/**
	 * 電文データ項目を登録します.
	 * @param	tag		項目名
	 * @param	data	設定するデータ
	 * @throws FieldNotFoundException
	 */
	public void setField( Tag tag, String data )
		throws FieldNotFoundException
	{
		Field field = getLayout().findField( tag );
		field.setValue( data );
	}

	/**
	 * 電文データ項目を登録します.
	 * @param	component	コンポーネント名
	 * @param	index	コンポーネントのインデックス
	 * @param	name	項目名
	 * @param	data	設定するデータ
	 * @throws FieldNotFoundException
	 */
	public void setField( String component, int index, String name, String data )
		throws FieldNotFoundException
	{
		Field field = getLayout().findField( component, index, name );
		field.setValue( data );
	}

	/**
	 * 電文データ項目を取得します.
	 * @param	name	項目名
	 * @return	項目値
	 * @throws	FieldNotFoundException
	 */
	public String getField( String name )
		throws FieldNotFoundException
	{
		Field field = getLayout().findField( name );
		return field.getValue();
	}

	/**
	 * 電文データ項目を取得します.
	 * @param	tag		項目名
	 * @return	項目値
	 * @throws	FieldNotFoundException
	 */
	public String getField( Tag tag )
		throws FieldNotFoundException
	{
		Field field = getLayout().findField( tag );
		return field.getValue();
	}

	/**
	 * 電文データ項目を取得します.
	 * @param	component	コンポーネント名
	 * @param	index	コンポーネントのインデックス
	 * @param	name	項目名
	 * @return	項目値
	 * @throws FieldNotFoundException
	 */
	public String getField( String component, int index, String name )
		throws FieldNotFoundException
	{
		Field field = getLayout().findField( component, index, name );
		return field.getValue();
	}

	/**
	 * 繰り返し項目の繰り返し数を取得します。
	 * @param	name	項目名
	 * @return	繰り返し数
	 * @throws	FIXException
	 */
	public int getGroupCount( String name )
		throws FIXException
	{
		Field field = getLayout().findField( name );
		if ( !( field instanceof Component )) {
			throw( new FIXException( "field name=[" + name + "] not component." ));
		}
		Component component = (Component )field;
		return component.getCount();
	}

	/**
	 * 繰り返し項目の繰り返し数を設定します。
	 * @param	name	項目名
	 * @param	count	繰り返し数
	 * @throws	FIXException
	 */
	public void setGroupCount( String name, int count )
		throws FIXException
	{
		Field field = getLayout().findField( name );
		if ( !( field instanceof Component )) {
			throw( new FIXException( "field name=[" + name + "] not component." ));
		}
		Component component = (Component )field;
		component.getCountField().setValue( String.valueOf( count ));
		component.setCount( count );
	}

	public byte[] getBytes() throws FIXException {
		try {
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			serialize( bos );
			byte[] ret = bos.toByteArray();
			return ret;
		} catch ( Exception ex ) {
			throw( new FIXException( "serialize error.", ex ));
		}
	}

	public void serialize( ByteArrayOutputStream bos )
		throws Exception
	{
		Component root = getLayout().getRootComponent();
		ByteArrayOutputStream bodyBos = new ByteArrayOutputStream();
		root.serialize( bodyBos );

		String var;
		Field field;
		field = getLayout().findField( Tag.BeginString );
		String ver = field.getValue();
		var = Tag.BeginString.getId() + "=" + ver;
		bos.write( var.getBytes() );
		bos.write( FIXMessage.SOH );

		bodyLength = String.valueOf( bodyBos.size() );
		var = Tag.BodyLength.getId() + "=" + bodyLength;
		bos.write( var.getBytes() );
		bos.write( FIXMessage.SOH );

		bodyBos.writeTo( bos );

		checkSum = calcCheckSum( bos );
		var = Tag.CheckSum.getId() + "=" + checkSum;
		bos.write( var.getBytes() );
		bos.write( FIXMessage.SOH );
		setLength( bos.size() );
	}

	String calcCheckSum( ByteArrayOutputStream bos ) {
		byte[] data = bos.toByteArray();
		long sum = 0;
		for ( int i = 0; i < data.length; i++ ) {
			sum += data[i];
		}
		DecimalFormat fmt = new DecimalFormat( "000" );
		String ret = fmt.format( sum % 256 );
		return ret;
	}

	public String toString() {
		Component root = getLayout().getRootComponent();
/*
		try {
			byte[] buf = getBytes();
		} catch ( FIXException ex ) {
		}
*/
		String ret = getMsgTypeName();
		ret += "" + root.toString( 0 );
		ret += "\nbodyLength=" + bodyLength;
		ret += " checkSum=" + checkSum;
		ret += " length=" + getLength();
		return ret;
	}

	byte[]	buffer;	// バッファ
	int	length;	// 電文長
	String	msgType;	// MsgType
	String	msgTypeName;	// MsgTypeName
	String	bodyLength;	// BodyLength
	String	checkSum;	// CheckSum
	String	description;	// メッセージ詳細

	/**
	* バッファを設定します。
	* @param	var		設定するデータ
	*/
	public void setBuffer( byte[] var ) {
		this.buffer = var;
	}
	/**
	* バッファを取得します。
	* @return	バッファ
	*/
	public byte[] getBuffer() {
		return buffer;
	}

	/**
	* 電文長を設定します。
	* @param	var		設定するデータ
	*/
	public void setLength( int var ) {
		this.length = var;
	}
	/**
	* 電文長を取得します。
	* @return	電文長
	*/
	public int getLength() {
		return length;
	}

	/**
	* MsgTypeを設定します。
	* @param	var		設定するデータ
	*/
	public void setMsgType( String var ) {
		this.msgType = var;
	}
	/**
	* MsgTypeを取得します。
	* @return	MsgType
	*/
	public String getMsgType() {
		return msgType;
	}

	/**
	* MsgTypeNameを設定します。
	* @param	var		設定するデータ
	*/
	public void setMsgTypeName( String var ) {
		this.msgTypeName = var;
	}
	/**
	* MsgTypeNameを取得します。
	* @return	MsgTypeName
	*/
	public String getMsgTypeName() {
		return msgTypeName;
	}

	/**
	* BodyLengthを設定します。
	* @param	var		設定するデータ
	*/
	public void setBodyLength( String var ) {
		this.bodyLength = var;
	}
	/**
	* BodyLengthを取得します。
	* @return	BodyLength
	*/
	public String getBodyLength() {
		return bodyLength;
	}

	/**
	* CheckSumを設定します。
	* @param	var		設定するデータ
	*/
	public void setCheckSum( String var ) {
		this.checkSum = var;
	}
	/**
	* CheckSumを取得します。
	* @return	CheckSum
	*/
	public String getCheckSum() {
		return checkSum;
	}

	/**
	* メッセージ詳細を設定します。
	* @param	var		設定するデータ
	*/
	public void setDescription( String var ) {
		this.description = var;
	}
	/**
	* メッセージ詳細を取得します。
	* @return	メッセージ詳細
	*/
	public String getDescription() {
		return description;
	}

}
