package	jp.co.sds_corp.fixconnect.transport;

import	java.io.ByteArrayOutputStream;
import	java.io.IOException;
import	java.io.InputStream;
import	java.io.OutputStream;
import	java.net.InetSocketAddress;
import	java.net.Socket;
import	java.net.SocketAddress;
import	java.net.SocketException;
import	java.net.SocketTimeoutException;
import	java.util.Date;
import	java.util.concurrent.atomic.AtomicBoolean;

import	jp.co.sds_corp.fixconnect.FIXException;
import	jp.co.sds_corp.fixconnect.message.FIXMessage;
import	jp.co.sds_corp.fixconnect.message.IMessageParser;
import	jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import	jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class FIXSocket
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	Socket	sock;
	OutputStream	sout;
	InputStream		sin;
	Thread	thr;
	byte[] recvBuf;
	ByteArrayOutputStream bos;
	IMessageParser	parser;	// メッセージ解析クラス

	TransportSupport transportSupport = new TransportSupport();
	ReceiverSupport recvSupport = new ReceiverSupport();

	public FIXSocket( Socket sock ) {
		bos = new ByteArrayOutputStream();
		this.sock = sock;
		setClosed( false );
	}

	public void start() throws FIXException {
		if ( isStarted()) {
			throw( new FIXException( "already started" ));
		}
		try {
			recvBuf = new byte[recvBufSize];
			sin = sock.getInputStream();
			sout = sock.getOutputStream();
			sock.setSoTimeout( getSocketTimeout() );

			thr = new Thread( new Receiver() );
			thr.setName( "aidw.fix.Receiver" );
			thr.start();
			setStarted( true );
			fireEventOccurred( TransportEvent.Type.CONNECT );
		} catch ( Exception ex ) {
			throw( new FIXException( "connect error", ex ));
		}
	}

	public void close() throws FIXException {
		if ( !isClosed()) {
			setClosed( true );
			if ( thr != null ) {
				logger.info( this, "closing Socket..." );
				try {
					thr.join();
				} catch ( Exception ex ) {
					// ignore
				}
			}
			try {
				sock.close();
			} catch ( IOException ex ) {
				throw( new FIXException( "socket close error", ex ));
			}
			logger.info( this, "socket closed. " + toString() );
		}
	}

	public synchronized void send( FIXMessage msg ) throws FIXException {
		byte[] buf = msg.getBytes();
		msg.setBuffer( buf );
		try {
			sout.write( buf );
		} catch ( IOException ex ) {
			throw( new FIXException( "message send error", ex ));
		}
	}

	public FIXMessage receive( int timeout ) throws Exception {
		long lastReceived = new Date().getTime();
		for ( ;; ) {
			if ( bos.size() > 0 ) {
				byte[] buf = bos.toByteArray();
				FIXMessage msg = null;
				int length = -1;
				try {
					length = parser.getMessageLength( buf );
				} catch ( FIXException ex ) {
					logger.error( this, "parse Exception ", ex );
					fireExceptionOccurred( ex );
				}
				if ( length < 0 ) {
					continue;
				}
				try {
					msg = parser.parse( buf );
					logger.trace( this, "message received "
						+ "MsgType=[" + msg.getMsgType() + "] "
						+ length + " bytes." );
					byte[] msgBuf = new byte[length];
					System.arraycopy( buf, 0, msgBuf, 0, length );
					msg.setBuffer( msgBuf );
				} catch ( FIXException ex ) {
					logger.error( this, "parse Exception ", ex );
					fireExceptionOccurred( ex );
				} finally {
					// parse済みバッファの消去
					bos.reset();
					if ( buf.length > length ) {
						bos.write( buf, length, buf.length - length );
					}
				}
				return msg;
			}

			int nRecv = -1;
			try {
				nRecv = sin.read( recvBuf, 0, recvBufSize );
			} catch ( SocketTimeoutException sex ) {
				if ( isClosed()) {
					return null;
				}
				long now = new Date().getTime();
				if ( timeout > 0 && ( now - lastReceived ) > timeout ) {
					return null;
				}
				continue;
			}
			if ( isClosed()) {
				return null;
			}
			if ( nRecv < 0 ) {
				throw( new SocketException( "connection closed."
											+ " (" + nRecv + ")" ));
			}
			lastReceived = new Date().getTime();
			bos.write( recvBuf, 0, nRecv );
		}
	}

	class Receiver implements Runnable
	{
		public void run() {
			for ( ;; ) {
				try {
					FIXMessage msg = receive( 0 );
					if ( msg != null ) {
						recvSupport.fireMessageReceived( msg );
					}
					if ( isClosed()) {
						break;
					}
				} catch ( Exception ex ) {
					if ( !isClosed()) {
						FIXException fex = new FIXException( "receive error", ex );
						fireExceptionOccurred( fex );
					}
					break;
				}
			}
			if ( !isClosed()) {
				try {
					sock.close();
				} catch ( IOException iex ) {
				}
				setClosed( true );
				logger.debug( this, "socket closed." );
			}
			fireEventOccurred( TransportEvent.Type.CLOSE );
			logger.debug( this, "receive thread terminated." );
		}
	}

	public void addEventListener( ITransportEventListener l ) {
		transportSupport.addEventListener( l );
	}

	public boolean removeEventListener( ITransportEventListener l ) {
		return transportSupport.removeEventListener( l );
	}

	public void fireEventOccurred( TransportEvent.Type type ) {
		transportSupport.fireEventOccurred( this, type );
	}

	public void fireEventOccurred( TransportEvent ev ) {
		transportSupport.fireEventOccurred( ev );
	}

	public void fireExceptionOccurred( FIXException ex ) {
		transportSupport.fireExceptionOccurred( ex );
	}

	public void addMessageReceiver( IMessageReceiver l ) {
		recvSupport.addMessageReceiver( l );
	}

	public boolean removeMessageReceiver( IMessageReceiver l ) {
		return recvSupport.removeMessageReceiver( l );
	}

	public void fireMessageReceived( MessageReceiveEvent ev ) {
		recvSupport.fireMessageReceived( ev );
	}

	public InetSocketAddress getRemoteAddress() {
		SocketAddress addr = sock.getRemoteSocketAddress();
		InetSocketAddress iaddr = (InetSocketAddress )addr;
		return iaddr;
	}

	/**
	* メッセージ解析クラスを設定します。
	* @param	var		設定するデータ
	*/
	public void setParser( IMessageParser var ) {
		this.parser = var;
	}
	/**
	* メッセージ解析クラスを取得します。
	* @return	メッセージ解析クラス
	*/
	public IMessageParser getParser() {
		return parser;
	}

	public String toString() {
		String ret = "FIXSocket";
		InetSocketAddress addr = getRemoteAddress();
		if ( addr != null ) {
			ret += " address=[" + addr + "]";
		}
		ret += " recvBufSize=" + getRecvBufSize();
		ret += " socketTimeout=" + getSocketTimeout();
		if ( isDebug()) {
			ret +=" DEBUG";
		}
		if ( isClosed()) {
			ret +=" CLOSED";
		}
		return ret;
	}

	boolean	ssl;	// SSLフラグ
	boolean	started;	// Socket起動フラグ
	// boolean	closed;	// Socketクローズフラグ
	AtomicBoolean closed = new AtomicBoolean( true );
	int	recvBufSize;	// 受信バッファサイズ
	int	socketTimeout;	// Socketタイムアウト
	boolean	debug;	// debugフラグ

	/**
	* SSLフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setSsl( boolean var ) {
		this.ssl = var;
	}
	/**
	* SSLフラグを取得します。
	* @return	SSLフラグ
	*/
	public boolean isSsl() {
		return ssl;
	}

	/**
	* Socket起動フラグを設定します。
	* @param	var		設定するデータ
	*/
	protected synchronized void setStarted( boolean var ) {
		this.started = var;
	}
	/**
	* Socket起動フラグを取得します。
	* @return	Socket起動フラグ
	*/
	public synchronized boolean isStarted() {
		return started;
	}

	/**
	* Socketクローズフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setClosed( boolean var ) {
		this.closed.set( var );
	}
	/**
	* Socketクローズフラグを取得します。
	* @return	Socketクローズフラグ
	*/
	public boolean isClosed() {
		return closed.get();
	}

	/**
	* 受信バッファサイズを設定します。
	* @param	var		設定するデータ
	*/
	public void setRecvBufSize( int var ) {
		this.recvBufSize = var;
	}
	/**
	* 受信バッファサイズを取得します。
	* @return	受信バッファサイズ
	*/
	public int getRecvBufSize() {
		return recvBufSize;
	}

	/**
	* Socketタイムアウトを設定します。
	* @param	var		設定するデータ
	*/
	public void setSocketTimeout( int var ) {
		this.socketTimeout = var;
	}
	/**
	* Socketタイムアウトを取得します。
	* @return	Socketタイムアウト
	*/
	public int getSocketTimeout() {
		return socketTimeout;
	}

	/**
	* debugフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setDebug( boolean var ) {
		this.debug = var;
	}
	/**
	* debugフラグを取得します。
	* @return	debugフラグ
	*/
	public boolean isDebug() {
		return debug;
	}

}
