package	jp.co.sds_corp.fixconnect.test;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;

import jp.co.sds_corp.fixconnect.FIXException;
import jp.co.sds_corp.fixconnect.message.FIXMessage;
import jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.fixconnect.transport.FIXSession;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class RateServer implements Runnable, IMessageReceiver
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	Thread	thread;
	boolean	stop;
	List<Request> listReq = new Vector<Request>();
	Map<String,Request> mapReq =
		new ConcurrentHashMap<String,Request>();
	FIXSession session;
	int	id;
	DecimalFormat df = new DecimalFormat( "000000" );
	SimpleDateFormat sdf = new SimpleDateFormat( "yyyyMMdd-HH:mm:ss.SSS" );
	Tenor tenor = new Tenor();

	enum ProductType {
		SPOT, OR, SWAP
	}

	class Request {
		FIXMessage	message;
		ProductType	productType;
		int count = 0;
		Request( FIXMessage message, ProductType productType ) {
			this.message = message;
			this.productType = productType;
		}
	}

	public RateServer( FIXSession session ) {
		this.session = session;
		session.addMessageReceiver( "R", this );
		session.addMessageReceiver( "Z", this );
	}

	public void onMessage( MessageReceiveEvent ev ) {
		FIXMessage req = ev.getMessage();
		String type = req.getMsgType();
		try {
			if ( type.equals( "R" )) {
				request( req );
			} else
			if ( type.equals( "Z" )) {
				cancel( req );
			}
		} catch ( Exception ex ) {
			logger.error( this, "response error", ex );
			try {
				reject( req, "error occurred" );
			} catch ( Exception ex2 ) {
			}
		}
	}

	void request( FIXMessage req ) throws Exception {
		String text = null;
		String requestId = req.getField( "QuoteReqID" );
		if ( requestId == null ) {
			text = "QuoteReqID[" + requestId + "] not defined.";
			reject( req, text );
			return;
		}
		Request org = mapReq.get( requestId );
		if ( org != null ) {
			text = "QuoteReqID[" + requestId + "] duplicated.";
			reject( req, text );
			return;
		}
		registerQuote( req );
	}

	void registerQuote( FIXMessage req ) throws FIXException {
		String requestId = req.getField( "QuoteReqID" );
		int count = req.getGroupCount( "Request" );
		for ( int i = 0; i < count; i++ ) {
			String ccy1 = req.getField( "Request", i, "Symbol" ).substring( 0, 3 );
			String ccy2 = req.getField( "Request", i, "Symbol" ).substring( 4 );
			if ( !ccy1.equals( "USD" ) || !ccy2.equals( "JPY" )) {
				logger.debug( this, "reject quote [" + ccy1 + "/" + ccy2 + "]" );
				String msg = "unknown symbol " + ccy1 + "/" + ccy2;
				try {
					reject( req, msg );
				} catch ( Exception ex ) {
				}
				return;
			}

			logger.debug( this, "register quote [" + ccy1 + "/" + ccy2 + "]" );
			String type = req.getField( "Request", i, "ProductType" );
			ProductType prodType = ProductType.valueOf( type );
			logger.debug( this, "request:" + requestId + " type=" + prodType );
			Request request = new Request( req, prodType );
			mapReq.put( requestId, request );
			listReq.add( request );
		}
	}

	void reject( FIXMessage req, String text ) throws Exception {
		FIXMessage res = getSession().getParser().createMessage( "AG" );
		String reqid = req.getField( "QuoteReqID" );
		res.setField( "QuoteReqID", reqid );
		res.setField( "QuoteRequestRejectReason", "99" );
		res.setField( "Text", text );
		getSession().send( res );
	}

	void cancel( FIXMessage req ) throws Exception {
		String text = null;
		String reqid = null;
		reqid = req.getField( "QuoteReqID" );
		logger.debug( this, "cancel:" + reqid );
		if ( reqid == null ) {
			text = "QuoteReqID not defined.";
			reject( req, text );
			return;
		}
		Request org = mapReq.remove( reqid );
		if ( org == null ) {
			text = "QuoteReqID[" + reqid + "] not found.";
			reject( req, text );
			return;
		}
		listReq.remove( org );
	}

	public void start() throws Exception {
		logger.info( this, "start server." );
		thread = new Thread( this );
		thread.setName( "aidw.fix.RateServer" );
		thread.start();
	}

	public void stop() throws Exception {
		logger.info( this, "stop server." );
		setStop( true );
		thread.join();
	}

	public void run() {
		for ( ;; ) {
			if ( isStop()) {
				break;
			}
			if ( getSession() == null ) {
				continue;
			}
			for ( Request req : listReq ) {
				if ( req.productType != ProductType.SPOT && req.count > 0 ) {
					continue;
				}
				try {
					FIXMessage res = buildResponse( req );
					getSession().send( res );
					req.count++;
				} catch ( Exception ex ) {
					logger.error( this, "send error", ex );
				}
			}
			try {
				Thread.sleep( 1000 );
			} catch ( InterruptedException ex ) {
			}
		}
		logger.info( this, "server stopped." );
	}

	synchronized String getQuoteId() {
		id++;
		String ret = "Q" + df.format( id );
		return ret;
	}

	FIXMessage buildResponse( Request request ) throws FIXException {
		FIXMessage req = request.message;
		FIXMessage res;
		res = getSession().getParser().createMessage( "S" );
		res.setField( "QuoteReqID", req.getField( "QuoteReqID" ));
		res.setField( "QuoteID", getQuoteId() );
		res.setField( "QuoteType", "0" );
		res.setField( "Account", req.getField( "Account" ));
		res.setField( "Symbol", req.getField( "Request[0].Symbol" ));
		res.setField( "ProductType", req.getField( "Request[0].ProductType" ));
		String product = req.getField( "Request[0].ProductType" );
		if ( product.equals( "SPOT" )) {
			buildSpot( req, res );
		} else
		if ( product.equals( "OR" )) {
			buildOutright( req, res );
		} else
		if ( product.equals( "SWAP" )) {
			buildSwap( req, res );
		}
		res.setField( "QuoteUnit", "1" );
		res.setField( "Currency", req.getField( "Request[0].Currency" ));

		return res;
	}

	void buildSpot( FIXMessage req, FIXMessage res ) throws FIXException {
		Date now = new Date();
		res.setField( "Tenor", "SP" );
		String date;
		date = tenor.getDate( "SP" );
		res.setField( "SettlDate", date );
		res.setField( "BidPx", "89.32" );
		res.setField( "OfferPx", "89.42" );
		res.setField( "BidSize", "10000000" );
		res.setField( "MinBidSize", "1000000" );
		res.setField( "OfferSize", "11000000" );
		res.setField( "MinOfferSize", "1100000" );
		res.setField( "Tier2BidPx", "88.82" );
		res.setField( "Tier2OfferPx", "89.92" );
		res.setField( "Tier2BidSize", "1000000" );
		res.setField( "Tier2MinBidSize", "100000" );
		res.setField( "Tier2OfferSize", "1100000" );
		res.setField( "Tier2MinOfferSize", "110000" );
		res.setField( "Tier3BidPx", "88.32" );
		res.setField( "Tier3OfferPx", "90.42" );
		res.setField( "Tier3BidSize", "100000" );
		res.setField( "Tier3MinBidSize", "10000" );
		res.setField( "Tier3OfferSize", "110000" );
		res.setField( "Tier3MinOfferSize", "11000" );
		res.setField( "TransactTime", sdf.format( now ));
	}

	void buildOutright( FIXMessage req, FIXMessage res ) throws FIXException {
		Date now = new Date();
		res.setField( "OrderQty", req.getField( "Request[0].OrderQty" ));
		res.setField( "Tenor", req.getField( "Request[0].Tenor" ));
		res.setField( "TenorTo", req.getField( "Request[0].TenorTo" ));
		String date;
		date = tenor.getDate( req.getField( "Request[0].Tenor" ));
		res.setField( "SettlDate",
			date != null ? date : req.getField( "Request[0].SettlDate" ));
		date = tenor.getDate( req.getField( "Request[0].TenorTo" ));
		res.setField( "SettlDateTo",
			date != null ? date : req.getField( "Request[0].SettlDateTo" ));
		res.setField( "BidPx", "89.3301" );
		res.setField( "OfferPx", "89.4302" );
		res.setField( "BidSpotRate", "88.32" );
		res.setField( "OfferSpotRate", "88.42" );
		res.setField( "BidForwardPoints", "1.01" );
		res.setField( "OfferForwardPoints", "1.02" );
		res.setField( "TransactTime", sdf.format( now ));
	}

	void buildSwap( FIXMessage req, FIXMessage res ) throws FIXException {
		Date now = new Date();
		res.setField( "OrderQty", req.getField( "Request[0].OrderQty" ));
		res.setField( "OrderQty2", req.getField( "Request[0].OrderQty2" ));
		res.setField( "Tenor", req.getField( "Request[0].Tenor" ));
		res.setField( "Tenor2", req.getField( "Request[0].Tenor2" ));
		String date;
		date = tenor.getDate( req.getField( "Request[0].Tenor" ));
		res.setField( "SettlDate",
			date != null ? date : req.getField( "Request[0].SettlDate" ));
		date = tenor.getDate( req.getField( "Request[0].Tenor2" ));
		res.setField( "SettlDate2",
			date != null ? date : req.getField( "Request[0].SettlDate2" ));
		res.setField( "BidPx", "89.3301" );
		res.setField( "OfferPx", "89.4302" );
		res.setField( "BidPx2", "89.3411" );
		res.setField( "OfferPx2", "89.4411" );
		res.setField( "BidSpotRate", "89.36" );
		res.setField( "OfferSpotRate", "89.38" );
		res.setField( "BidForwardPoints", "1.01" );
		res.setField( "OfferForwardPoints", "1.02" );
		res.setField( "BidForwardPoints2", "2.11" );
		res.setField( "OfferForwardPoints2", "2.12" );
		res.setField( "BidSwapPoints", "1.11" );
		res.setField( "OfferSwapPoints", "1.12" );
		res.setField( "TransactTime", sdf.format( now ));
	}

	public synchronized void setStop( boolean var ) {
		this.stop = var;
	}
	public synchronized boolean isStop() {
		return this.stop;
	}

	public synchronized void setSession( FIXSession var ) {
		this.session = var;
	}
	public synchronized FIXSession getSession() {
		return this.session;
	}

}
